from django import forms

from alert.models import AlertPreference
from alert.utils import ALERT_TYPES



class AlertPreferenceForm(forms.Form):
    
    def __init__(self, user, *args, **kwargs):
        self.user = user
        self.prefs = AlertPreference.objects.get_user_prefs(user).items()
        
        super(AlertPreferenceForm, self).__init__(*args, **kwargs)
        
        for ((notice_type, backend), pref) in self.prefs:
            attr = self._field_id(notice_type, backend)
            alert = ALERT_TYPES[notice_type]
            self.fields[attr] = forms.BooleanField(
                                                   label=alert.title, 
                                                   help_text=alert.description,
                                                   initial=pref,
                                                   required=False
                                                   )
        
        
    def save(self):
        for (notice_type, backend), pref in self.prefs:
            attr = self._field_id(notice_type, backend)

            alert_pref, created = AlertPreference.objects.get_or_create(
                                            user=self.user,
                                            alert_type=notice_type,
                                            backend=backend
                                            )
            alert_pref.preference = self.cleaned_data[attr]
            alert_pref.save()
            
            
    def _field_id(self, notice_type, backend):
        return "%s__%s" % (notice_type, backend)


from datetime import datetime

from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.db import models

from alert.utils import ALERT_TYPE_CHOICES, ALERT_BACKEND_CHOICES, ALERT_TYPES, ALERT_BACKENDS
from alert.managers import AlertManager, PendingAlertManager, AlertPrefsManager
from alert.exceptions import CouldNotSendError
from alert.signals import alert_sent


class Alert(models.Model):
    user = models.ForeignKey(User)
    backend = models.CharField(max_length=20, default='EmailBackend', choices=ALERT_BACKEND_CHOICES)
    alert_type = models.CharField(max_length=25, choices=ALERT_TYPE_CHOICES)
    
    title = models.CharField(max_length=250, default=lambda: "%s alert" % Site.objects.get_current().name)
    body = models.TextField()
    
    when = models.DateTimeField(default=datetime.now)
    created = models.DateTimeField(default=datetime.now)
    last_attempt = models.DateTimeField(blank=True, null=True)
    
    is_sent = models.BooleanField(default=False)
    failed = models.BooleanField(default=False)
    
    objects = AlertManager()
    pending = PendingAlertManager()
    
    
    def send(self):
        backend = Alert._get_backend(self.backend)
        try:
            backend.send(self)
            self.is_sent = True
            self.failed = False
            alert_sent.send(sender=ALERT_TYPES[self.alert_type], alert=self)
            
        except CouldNotSendError:
            self.failed = True
        
        self.last_attempt = datetime.now()
        self.save()
        
        
    @classmethod
    def _get_backend(cls, backend):
        return ALERT_BACKENDS[backend]



class AlertPreference(models.Model):
    user = models.ForeignKey(User)
    alert_type = models.CharField(max_length=25, choices=ALERT_TYPE_CHOICES)
    backend = models.CharField(max_length=25, choices=ALERT_BACKEND_CHOICES)
    
    preference = models.BooleanField()
    
    objects = AlertPrefsManager()
    
    class Meta:
        unique_together = ('user', 'alert_type', 'backend')
   

import alert.backends #@UnusedImport


from datetime import datetime
from django.core.mail import send_mail
from django.conf import settings

from alert.utils import BaseAlertBackend
from alert.exceptions import CouldNotSendError

class EmailBackend(BaseAlertBackend):
    title = "Email"
    
    def send(self, alert):
        recipient = alert.user.email
        if not recipient: raise CouldNotSendError
        
        send_mail(alert.title, alert.body, settings.DEFAULT_FROM_EMAIL, [recipient])
        
    


from datetime import datetime
from collections import defaultdict
from django.db.models import Manager
from alert.utils import ALERT_TYPES, ALERT_BACKENDS
from django.contrib.auth.models import User


class AlertManager(Manager): pass


class PendingAlertManager(AlertManager):
    
    def get_queryset(self, *args, **kwargs):
        qs = super(PendingAlertManager, self).get_queryset(*args, **kwargs)
        return qs.filter(is_sent=False, failed=False, when__gte=datetime.now())
    
    
    
    
class AlertPrefsManager(Manager):
    
    def get_user_prefs(self, user):
        alert_prefs = self.get_query_set().filter(user=user)
        
        prefs = {}
        for pref in alert_prefs:
            prefs[pref.alert_type, pref.backend] = pref.preference
        
        for notice_type in ALERT_TYPES.values():
            for backend in ALERT_BACKENDS.values():
                if (notice_type.id, backend.id) not in prefs:
                    default_pref = notice_type.get_default(backend.id)
                    prefs[notice_type.id, backend.id] = default_pref
        return prefs
    
                
    def get_recipients_for_notice(self, notice_type, users):
        if isinstance(notice_type, basestring):
            notice_type = ALERT_TYPES[notice_type]
        
        alert_prefs = self.get_query_set().filter(alert_type=notice_type.id).filter(user__in=users)
        
        prefs = {}
        for pref in alert_prefs:
            prefs[pref.user_id, pref.backend] = pref.preference
        
        for user in users:
            for backend in ALERT_BACKENDS.values():
                if (user.id, backend.id) not in prefs:
                    prefs[user.id, backend.id] = notice_type.get_default(backend.id)
        
        return tuple((User.objects.get(id=user_id), ALERT_BACKENDS[backend_id]) for ((user_id, backend_id), pref) in prefs.items() if pref)
    



from django.contrib import admin
from models import Alert, AlertPreference

class AlertAdmin(admin.ModelAdmin):
    list_display = ('user', 'backend', 'title', 'created', 'failed', 'is_sent')

class AlertPrefAdmin(admin.ModelAdmin):
    list_display = ("user", 'alert_type', "backend", 'preference')

admin.site.register(Alert, AlertAdmin)
admin.site.register(AlertPreference, AlertPrefAdmin)


from alert.exceptions import AlertIDAlreadyInUse, AlertBackendIDAlreadyInUse
from django.template.loader import render_to_string, find_template
from django.contrib.sites.models import Site
from django.template import TemplateDoesNotExist

ALERT_TYPES = {}
ALERT_BACKENDS = {}

ALERT_TYPE_CHOICES = [] 
ALERT_BACKEND_CHOICES = []



class AlertMeta(type):

    def __new__(cls, name, bases, attrs):
        new_alert = super(AlertMeta, cls).__new__(cls, name, bases, attrs)
        
        # If this isn't a subclass of BaseAlert, don't do anything special.
        parents = [b for b in bases if isinstance(b, AlertMeta)]
        if not parents: 
            return new_alert
        
        new_alert.id = getattr(new_alert, 'id', name)
        
        if new_alert.id in ALERT_TYPES.keys():
            raise AlertIDAlreadyInUse("The alert ID, \"%s\" was delared more than once" % new_alert.id)
        
        ALERT_TYPES[new_alert.id] = new_alert()
        ALERT_TYPE_CHOICES.append((new_alert.id, new_alert.title))
        
        return new_alert



class BaseAlert(object):
    __metaclass__ = AlertMeta
    
    default = False
    sender = None
    template_filetype = "txt"
    
    
    def __init__(self):
        kwargs = {}
        if self.sender:
            kwargs['sender'] = self.sender
        
        self.signal.connect(self.signal_handler, **kwargs)
    
    
    def signal_handler(self, **kwargs):
    
        if self.before(**kwargs) is False: 
            return

        from alert.models import AlertPreference
        from alert.models import Alert
        
        users = self.get_applicable_users(**kwargs)
        site = Site.objects.get_current()
        
        for user, backend in AlertPreference.objects.get_recipients_for_notice(self.id, users):
            context = self.get_template_context(BACKEND=backend, USER=user, SITE=site, **kwargs)
            template_kwargs = {'backend': backend, 'context': context } 
            Alert.objects.create(
                                 user=user, 
                                 backend=backend.id,
                                 alert_type=self.id,
                                 title=self.get_title(**template_kwargs),
                                 body=self.get_body(**template_kwargs)
                                 )
    
    
    def before(self, **kwargs):
        pass
    
    
    def get_applicable_users(self, instance, **kwargs):
        return [instance.user]
    
    
    def get_template_context(self, **kwargs):
        return kwargs


    def _get_template(self, backend, part, filetype='txt'):
        template = "alerts/%s/%s/%s.%s" % (self.id, backend.id, part, filetype)
        try:
            find_template(template)
            return template
        except TemplateDoesNotExist:
            pass
        
        template = "alerts/%s/%s.%s" % (self.id, part, filetype)
        find_template(template)
        
        return template
        
    
    def get_title_template(self, backend, context):
        return self._get_template(backend, 'title', self.template_filetype)
    
    
    def get_body_template(self, backend, context):
        return self._get_template(backend, 'body', self.template_filetype)
    
    
    def get_title(self, backend, context):
        template = self.get_title_template(backend, context)
        return render_to_string(template, context)
    
    
    def get_body(self, backend, context):
        template = self.get_body_template(backend, context)
        return render_to_string(template, context)
    
    
    def get_default(self, backend):
        if isinstance(self.default, bool): 
            return self.default
        return self.default[backend]
    
    

class AlertBackendMeta(type):

    def __new__(cls, name, bases, attrs):
        new_alert_backend = super(AlertBackendMeta, cls).__new__(cls, name, bases, attrs)
        
        # If this isn't a subclass of BaseAlert, don't do anything special.
        parents = [b for b in bases if isinstance(b, AlertBackendMeta)]
        if not parents: 
            return new_alert_backend
        
        new_alert_backend.id = getattr(new_alert_backend, 'id', name)
        
        if new_alert_backend.id in ALERT_BACKENDS.keys(): 
            raise AlertBackendIDAlreadyInUse("The alert ID, \"%s\" was delared more than once" % new_alert_backend.id)
        
        ALERT_BACKENDS[new_alert_backend.id] = new_alert_backend()
        ALERT_BACKEND_CHOICES.append((new_alert_backend.id, new_alert_backend.title))
        
        return new_alert_backend



class BaseAlertBackend(object):
    __metaclass__ = AlertBackendMeta
    
    
    def mass_send(self, alerts):
        try:
            [self.send(alert) for alert in alerts]
        except TypeError:
            self.send(alerts)
    
    


from alert.utils import BaseAlert
from django.contrib.auth.models import User



class WelcomeAlert(BaseAlert):
    title = 'Welcome Users'
    description = 'When news breaks, be the first to react and profit.'
    
    signal = new_feature
    default = False
    
    def get_applicable_users(self, **kwargs):
        return User.objects.all()


class NewTopic(BaseAlert):
    title = 'New Topic Alert'
    description = 'Get notified as soon as new topics go live.'
    
    signal = market_opened
    default = False
    
    def get_applicable_users(self, **kwargs):
        return User.objects.all()


class CashOut(BaseAlert):
    title = 'Topic cash-out alert'
    description = 'Get notified of your results when a topic is judged.'
    
    signal = market_cashed_out
    default = True
    
    def get_applicable_users(self, cashout_orders, **kwargs):
        profiles = set(order.userprofile for order in cashout_orders)
        return set(prof.user for prof in profiles)
    
    def get_template_context(self, BACKEND, USER, cashout_orders, **kwargs):
        order = [o for o in cashout_orders if o.userprofile.user_id == USER.id]
        locals().update(kwargs)
        return locals()


class TopicClosing(BaseAlert):
    title = 'Closing soon alert'
    description = 'Get notified when topics are closing.'
    
    signal = market_closing_soon
    default = True
    
    def get_applicable_users(self, instance, days_left, **kwargs):
        if days_left != 1:
            return User.objects.none()
        market = instance
        return User.objects.filter(userprofile__order__market=market).distinct()

    
    